package transactionFeature;

import com.sleepycat.je.DatabaseException;
import com.sleepycat.je.Environment;
import com.sleepycat.je.TransactionConfig;
import com.sleepycat.je.dbi.EnvironmentImpl;
import com.sleepycat.je.txn.Txn;
import com.sleepycat.je.txn.TxnManager;

privileged public aspect BeginAndEndTransaction {
	/**
	 * Javadoc for this public method is generated via the doc templates in the
	 * doc_src directory.
	 */
	public Transaction Environment.beginTransaction(Transaction parent,
			TransactionConfig txnConfig) throws DatabaseException {

		checkHandleIsValid();
		checkEnv();

		if (!environmentImpl.isTransactional()) {
			throw new DatabaseException(
					"Transactions can not be used in a non-transactional "
							+ "environment");
		}

		if (txnConfig != null
				&& ((txnConfig.getSerializableIsolation() && txnConfig
						.getReadUncommitted())
						|| (txnConfig.getSerializableIsolation() && txnConfig
								.getReadCommitted()) || (txnConfig
						.getReadUncommitted() && txnConfig.getReadCommitted()))) {
			throw new IllegalArgumentException(
					"Only one may be specified: SerializableIsolation, "
							+ "ReadCommitted or ReadUncommitted");
		}

		/*
		 * Apply txn config defaults. We don't need to clone unless we have to
		 * apply the env default, since we don't hold onto a txn config
		 * reference.
		 */
		TransactionConfig useConfig = null;
		if (txnConfig == null) {
			useConfig = defaultTxnConfig;
		} else {
			if (defaultTxnConfig.getNoSync()
					|| defaultTxnConfig.getWriteNoSync()) {

				/*
				 * The environment sync settings have been set, check if any
				 * were set in the user's txn config. If none were set in the
				 * user's config, apply the environment defaults
				 */
				if (!txnConfig.getNoSync() && !txnConfig.getSync()
						&& !txnConfig.getWriteNoSync()) {
					useConfig = txnConfig.cloneConfig();
					if (defaultTxnConfig.getWriteNoSync()) {
						useConfig.setWriteNoSync(true);
					} else {
						useConfig.setNoSync(true);
					}
				}
			}

			/* Apply isolation level default. */
			if (!txnConfig.getSerializableIsolation()
					&& !txnConfig.getReadCommitted()
					&& !txnConfig.getReadUncommitted()) {
				if (defaultTxnConfig.getSerializableIsolation()) {
					if (useConfig == null) {
						useConfig = txnConfig.cloneConfig();
					}
					useConfig.setSerializableIsolation(true);
				} else if (defaultTxnConfig.getReadCommitted()) {
					if (useConfig == null) {
						useConfig = txnConfig.cloneConfig();
					}
					useConfig.setReadCommitted(true);
				}
			}

			/* No environment level defaults applied. */
			if (useConfig == null) {
				useConfig = txnConfig;
			}
		}

		Transaction txn = new Transaction(this, environmentImpl.txnBegin(
				parent, useConfig));
		addReferringHandle(txn);
		return txn;
	}
	
	/**
	 * Transactional services.
	 */
	public Txn EnvironmentImpl.txnBegin(Transaction parent, TransactionConfig txnConfig)
			throws DatabaseException {

		if (!isTransactional) {
			throw new DatabaseException("beginTransaction called, "
					+ " but Environment was not opened "
					+ "with transactional cpabilities");
		}

		return txnManager.txnBegin(parent, txnConfig);
	}
	
	/**
	 * Create a new transaction.
	 * 
	 * @param parent
	 *            for nested transactions, not yet supported
	 * @param txnConfig
	 *            specifies txn attributes
	 * @return the new txn
	 */
	public Txn TxnManager.txnBegin(Transaction parent, TransactionConfig txnConfig)
			throws DatabaseException {

		if (parent != null) {
			throw new DatabaseException(
					"Nested transactions are not supported yet.");
		}

		return new Txn(env, txnConfig);
	}

}
