package transactionFeature;

import com.sleepycat.je.Cursor;
import com.sleepycat.je.CursorConfig;
import com.sleepycat.je.Database;
import com.sleepycat.je.DatabaseConfig;
import com.sleepycat.je.DatabaseException;
import com.sleepycat.je.Environment;
import com.sleepycat.je.EnvironmentConfig;
import com.sleepycat.je.JoinConfig;
import com.sleepycat.je.JoinCursor;
import com.sleepycat.je.config.BooleanConfigParam;
import com.sleepycat.je.config.EnvironmentParams;
import com.sleepycat.je.dbi.DatabaseId;
import com.sleepycat.je.dbi.DatabaseImpl;
import com.sleepycat.je.dbi.DbConfigManager;
import com.sleepycat.je.dbi.EnvironmentImpl;
import com.sleepycat.je.txn.BasicLocker;
import com.sleepycat.je.txn.Locker;

privileged public aspect TransactionalProperties {

	/**
	 * Database
	 */

	/**
	 * Equivalent to getConfig().getTransactional() but cheaper.
	 */
	boolean Database.isTransactional() throws DatabaseException {
		return databaseImpl.isTransactional();
	}

	after(Database db) : 
		execution(void Database.initExisting(Environment, Locker, DatabaseImpl, DatabaseConfig)) 
		&& this(db) && within(Database) 
	{
		db.configuration.setTransactional(db.databaseImpl.isTransactional());
	}

	after(Database db, DatabaseConfig config, DatabaseImpl databaseImpl) throws DatabaseException : 
		execution(void Database.validateConfigAgainstExistingDb(DatabaseConfig, DatabaseImpl)) && 
		args(config,databaseImpl) && this(db) {
		/*
		 * The transactional property is kept constant while any handles are
		 * open, and set when the first handle is opened. It does not need to be
		 * specified if the useExistingConfig property is set.
		 */
		if (databaseImpl.hasOpenHandles()) {
			if (!config.getUseExistingConfig()) {
				if (config.getTransactional() != databaseImpl.isTransactional()) {
					throw new DatabaseException(
							"You can't open a Database with a transactional "
									+ "configuration of "
									+ config.getTransactional()
									+ " if the underlying database was created with a "
									+ "transactional configuration of "
									+ databaseImpl.isTransactional() + ".");
				}
			}
		} else {
			databaseImpl.setTransactional(config.getTransactional());
		}
	}

	/*
	 * Check that all cursors use the same locker, if any cursor is
	 * transactional. And if non-transactional, that all databases are in
	 * the same environment.
	 */
	JoinCursor around (Cursor[] cursors, Database db, Locker locker) throws DatabaseException : 
		call(JoinCursor.new(Locker, Database, Cursor[], JoinConfig)) && 
		args(locker,*,cursors,*) && 
		withincode(JoinCursor Database.join(Cursor[], JoinConfig)) && this(db)
	{
		locker = cursors[0].getCursorImpl().getLocker();
		if (!locker.isTransactional()) {
			for (int i = 1; i < cursors.length; i += 1) {
				Locker locker2 = cursors[i].getCursorImpl().getLocker();
				if (locker2.isTransactional()) {
					throw new IllegalArgumentException(
							"All cursors must use the same transaction.");
				}
			}
			locker = null; /* Don't reuse a non-transactional locker. */
		} else {
			for (int i = 1; i < cursors.length; i += 1) {
				Locker locker2 = cursors[i].getCursorImpl().getLocker();
				if (locker.getTxnLocker() != locker2.getTxnLocker()) {
					throw new IllegalArgumentException(
							"All cursors must use the same transaction.");
				}
			}
		}
		
		return proceed(cursors,db,locker);
	}

	/**
	 * DatabaseImpl
	 */
	before(DatabaseImpl dbimpl, DatabaseConfig dbConfig) : 
		execution(DatabaseImpl.new(String, DatabaseId, EnvironmentImpl, DatabaseConfig)) && 
		args(String, DatabaseId, EnvironmentImpl, dbConfig) && this(dbimpl)
	{
		dbimpl.transactional = dbConfig.getTransactional();
	}

	private boolean DatabaseImpl.transactional;// All open handles are

	// transactional

	/**
	 * @return true if this database is transactional.
	 */
	public boolean DatabaseImpl.isTransactional() {
		return transactional;
	}

	/**
	 * Sets the transactional property for the first opened handle.
	 */
	public void DatabaseImpl.setTransactional(boolean transactional) {
		this.transactional = transactional;
	}

	/**
	 * EnvironmentImpl
	 */
	private boolean EnvironmentImpl.isTransactional; // true if env opened

	// with DB_INIT_TRANS
	public boolean EnvironmentImpl.isTransactional() {
		return isTransactional;
	}

	void around(EnvironmentImpl envimpl, DbConfigManager configManager)
			throws DatabaseException : 
		call(void EnvironmentImpl.hook_readProperties(DbConfigManager)) && 
		args(configManager) && this(envimpl)
	{
		envimpl.isTransactional = configManager
				.getBoolean(EnvironmentParams.ENV_INIT_TXN);
		if (envimpl.isTransactional && envimpl.isNoLocking) {
			if (EnvironmentImpl.TEST_NO_LOCKING_MODE) {
				envimpl.isNoLocking = !envimpl.isTransactional;
			} else {
				throw new IllegalArgumentException(
						"Can't set 'je.env.isNoLocking' and "
								+ "'je.env.isTransactional';");
			}
		}
	}

	/*
	 * ChK_ not made final because of technical reaons. the field is initialized
	 * before another resulting an a null pointer exception
	 */
	public static BooleanConfigParam EnvironmentParams.ENV_INIT_TXN = null;

	after() : staticinitialization(EnvironmentParams) {
		EnvironmentParams.ENV_INIT_TXN = new BooleanConfigParam(
				"je.env.isTransactional", false, // default
				false,// mutable
				"# If true, create the environment w/ transactions.");
	}

	/**
	 * DatabaseConfig
	 */
	private boolean DatabaseConfig.transactional = false;

	/**
	 * Javadoc for this public method is generated via the doc templates in the
	 * doc_src directory.
	 */
	public void DatabaseConfig.setTransactional(boolean transactional) {
		this.transactional = transactional;
	}

	/**
	 * Javadoc for this public method is generated via the doc templates in the
	 * doc_src directory.
	 */
	public boolean DatabaseConfig.getTransactional() {
		return transactional;
	}

	/**
	 * EnvironmentConfig
	 */
	/**
	 * Javadoc for this public method is generated via the doc templates in the
	 * doc_src directory.
	 */
	public void EnvironmentConfig.setTransactional(boolean transactional) {
		setVal(EnvironmentParams.ENV_INIT_TXN, Boolean.toString(transactional));
	}

	/**
	 * Javadoc for this public method is generated via the doc templates in the
	 * doc_src directory.
	 */
	public boolean EnvironmentConfig.getTransactional() {
		String val = getVal(EnvironmentParams.ENV_INIT_TXN);
		return (Boolean.valueOf(val)).booleanValue();
	}

	/**
	 * Locker
	 */
	/**
	 * Returns whether this locker is transactional.
	 */
	public abstract boolean Locker.isTransactional();

	/**
	 * BasicLocker
	 */
	/**
	 * Is never transactional.
	 */
	public boolean BasicLocker.isTransactional() {
		return false;
	}

	/**
	 * misc checks
	 */
	before(Environment env, DatabaseConfig dbConfig, String databaseName)
			throws DatabaseException:
		execution(void Environment.validateDbConfigAgainstEnv(DatabaseConfig, String)) && 
		this(env) && args(dbConfig,databaseName) 
    {
		/*
		 * Check operation's transactional status against the Environment
		 */
		if (dbConfig.getTransactional()
				&& !(env.environmentImpl.isTransactional())) {
			throw new DatabaseException("Attempted to open Database "
					+ databaseName
					+ " transactionally, but parent Environment is"
					+ " not transactional");
		}
	}

	boolean around(boolean isWritable, Locker locker, DatabaseImpl dbImpl) : 
		call(boolean Cursor.hook_getUpdateOperationsProhibited(boolean)) && args(isWritable) &&
		cflow(execution(void Cursor.init(Database, DatabaseImpl, Locker, boolean, CursorConfig)) &&
				args(Database,dbImpl,locker, boolean, CursorConfig))
	{
		return (dbImpl.isTransactional() && !locker.isTransactional())
				|| !isWritable;
	}
}
